/*
 * HttpD.c
 *
 *      Author: Trax
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <avr/pgmspace.h>
#include "HttpD.h"
#include "../uip/uip.h"
#include "../../Ports.h"
#include "../../Options.h"
#include "../../Screens.h"

void Reboot();

void Save(int n);

char g_Buffer[128];

int ParseHeader(char* pBuff, int iLen, tTrans* Trans);
int MakeHttpData(char* pBuff, int nMSS, tTrans* Trans);

#define STATE_CONNECTED		0
#define STATE_SENDHEADER	1
#define STATE_HEADERSENT	2
#define STATE_SENDDATA		3
#define STATE_DATASENT		4

uint8_t strcopy_P(char** pBuff, int* nMSS, int* iPos, const prog_char* pSrc);

static const prog_char g_HttpHeader[] =
				"HTTP/1.1 200 OK\r\n"
				"Content-Type:text/html\r\n"
				"Connection:close\r\n"
				"Transfer-Encoding:chunked\r\n"
				"\r\n";

void HttpDInit()
{
	//Start listening on our port
	uip_listen(HTONS(PORT_HTTPD));
}

void SendHttpData(char* DATA, int MSS, tTrans* Trans)
{
	uint16_t nBufSize = MakeHttpData(DATA+5, MSS-7, Trans);
	if (nBufSize == 0)
	{
		Trans->State = STATE_DATASENT;
		memcpy_P(uip_appdata,PSTR("0\r\n\r\n"),5);
		nBufSize = 5;
	}
	else
	{
		memcpy_P(uip_appdata,PSTR("   \r\n"),5);
		Trans->n = sprintf_P(Trans->Buff,PSTR("%x"),nBufSize);
		memcpy(uip_appdata,Trans->Buff,Trans->n);
		memcpy_P(DATA+5+nBufSize,PSTR("\r\n"),2);
		Trans->n = 0;
		nBufSize += 7;
	}
	uip_send(uip_appdata, nBufSize);
}

void HttpDCall(	char* pBuffer,
				int nBytes,
				tTrans* pSocket)
{
	if (uip_connected())
	{
		pSocket->State = STATE_CONNECTED;
		pSocket->Index = 0;
		pSocket->h=pSocket->i=pSocket->j=pSocket->k=pSocket->l=pSocket->m=0;
	}
	else if (uip_newdata() || uip_acked())
	{
		if (pSocket->State == STATE_CONNECTED)
		{
			if(ParseHeader(pBuffer,nBytes,pSocket) == 0)
				pSocket->State = STATE_SENDHEADER;
		}

		if (pSocket->State == STATE_SENDHEADER)
		{
			pSocket->State = STATE_HEADERSENT;
			strcpy_P(uip_appdata, g_HttpHeader);
			uip_send(uip_appdata,strlen_P(g_HttpHeader));
			return;
		}

		if(pSocket->State == STATE_HEADERSENT)
		{
			pSocket->State = STATE_SENDDATA;
		}

		if (pSocket->State == STATE_SENDDATA)
		{
			// backup state
			pSocket->h_ = pSocket->h;
			pSocket->i_ = pSocket->i;
			pSocket->j_ = pSocket->j;
			pSocket->k_ = pSocket->k;
			pSocket->l_ = pSocket->l;
			pSocket->m_ = pSocket->m;
			pSocket->n_ = pSocket->n;

			SendHttpData(uip_appdata, uip_mss(), pSocket);
		}
		else if(pSocket->State == STATE_DATASENT)
			uip_close();
	}
	else if (uip_rexmit())
	{
		if (pSocket->State == STATE_HEADERSENT)
		{
			strcpy_P(uip_appdata, g_HttpHeader);
			uip_send(uip_appdata,strlen_P(g_HttpHeader));
		}
		else if (pSocket->State == STATE_SENDDATA)
		{
			// restore state
			pSocket->h = pSocket->h_;
			pSocket->i = pSocket->i_;
			pSocket->j = pSocket->j_;
			pSocket->k = pSocket->k_;
			pSocket->l = pSocket->l_;
			pSocket->m = pSocket->m_;
			pSocket->n = pSocket->n_;

			SendHttpData(uip_appdata, uip_mss(), pSocket);
		}
		else if(pSocket->State == STATE_DATASENT)
		{
			memcpy_P(uip_appdata,PSTR("0\r\n\r\n"),5);
			uip_send(uip_appdata, 5);
		}
	}
}



// Combo box Formating functions

uint8_t strcopy_C(char** pBuff, int* nMSS, int* iPos, const char* pSrc)
{
	char Char;
	while((Char = pSrc[*iPos]) != '\0')
	{
		**pBuff = Char;
		(*pBuff)++;
		(*iPos)++;
		if(--*nMSS == 0)
			return 1;
	}
	*iPos = 0;
	return 0;
}

uint8_t strcopy_P(char** pBuff, int* nMSS, int* iPos, const prog_char* pSrc)
{
	char Char;
	while ((Char = pgm_read_byte(pSrc+*iPos)) != '\0')
	{
		**pBuff = Char;
		(*pBuff)++;
		(*iPos)++;
		if(--*nMSS == 0)
			return 1;
	}
	*iPos = 0;
	return 0;
}

uint8_t strprint(char** pBuff, int* nMSS, int* iPos, const prog_char * _Format, ...)
{
	va_list argptr;
	va_start(argptr, _Format);
#ifdef WIN32
	_vsnprintf(g_Buffer, ARRSIZE(g_Buffer), _Format, argptr);
#else
	vsnprintf_P(g_Buffer, ARRSIZE(g_Buffer), _Format, argptr);
#endif
	va_end(argptr);

	return strcopy_C(pBuff, nMSS, iPos, g_Buffer);
}

#ifdef WIN32
#define strncmp_PP(_L,_R,_n)        strncmp(_L,_R,_n)
#define strcmp_PP(_L,_R)       		strcmp(_L,_R)
#define strcmp_P(_L,_R)       		strcmp(_L,_R)
#define strcpy_P(_L,_R)       		strcpy(_L,_R)
#define strlen_P(_L)       			strlen(_L)
#else
#define strncmp_PP(_L,_R,_n)       strncmp_P(_R,_L,_n)
/*
#define strncmp_PP(_L,_R,_n)       strncmp_P(_L,PSTR(__R)) // length is always 2
int _strncmp_PP(const prog_char * L,const prog_char * R)
{
	if(pgm_read_byte(L) == pgm_read_byte(R) && pgm_read_byte(L+1) == pgm_read_byte(R+1))
		return 0;
	else
		return 1;
}
*/
#endif

// Common HTML Header:
static const prog_char g_HtmlHeaderBegin[] =
				"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\">\r\n"
				"<html><head>\r\n"
				"<title>AVR Smart WebSrv by Trax Xavier</title>\r\n";


static const prog_char g_HtmlHeaderStyle[] =
				"<html><head>\r\n"
				"<title>AVR Smart WebSrv by Trax Xavier</title>\r\n"
				"<style type='text/css'>\r\n"
				".s1 { background-color: #FF7777; }\r\n"
				".s2 { background-color: #77FF77; }\r\n"
				".s3 { background-color: #7777FF; }\r\n"
				".s4 { background-color: #FFFF77; }\r\n"
				".s5 { background-color: #FFFFFF; }\r\n"
				"td { text-align: center; }\r\n"
				"</style>\r\n";

static const prog_char g_HtmlHeaderScript[] =
				"<script type='text/javascript'>\r\n"
				"function genCombo(list , name, sel , submit) {\r\n"
				"   var s = '<select name=\\'' + name + '\\'';\r\n"
				"   if(submit != '')\r\n"
				"      s += 'onchange=\\'' + submit + '.submit();\\'';\r\n"
				"   s += '>';\r\n"
				"   for (var i = 0; i < list.length; i++) {\r\n"
				"      s += '<option value=\\'' + list[i] + '\\'';\r\n"
				"      if (list[i] == sel)\r\n"
				"         s += ' selected=\\'selected\\'';\r\n"
				"      s += '>' + list[i] + '</option>';\r\n"
				"   }\r\n"
				"   document.write(s + '</select>');\r\n"
				"}\r\n"
				"</script>\r\n";

static const prog_char g_HtmlHeaderEnd[] =
				"</head><body>\r\n";

static const prog_char g_HtmlFooter[] =
				"</body></html>\r\n";


// Combo box FOrmating functions

static const prog_char g_HtmlOptsBegin[] =
				"<script type='text/javascript'> var %s = new Array('None'";

static const prog_char g_HtmlOpts[] =
				", '%s'";

static const prog_char g_HtmlOptsEnd[] =
				");</script>\r\n";

static const prog_char g_HtmlCBNo[] =
				"<noscript><input name='%s%d' value='%s' maxlength='8' size='2' type='text'></noscript>";

static const prog_char g_HtmlCBScript[] =
				"<script type='text/javascript'>genCombo(%s, '%s%d', '%s', '%s');</script>";

int PrintPortsList(char** pBuff, int* nMSS, int* iStage, int* iStep, int* iPos, char* pName, int Type)
{
	if(*iStage == 0)
	{
		if(strprint(pBuff, nMSS, iPos, g_HtmlOptsBegin, pName))
			return 1;
		*iStage = 1;
	}
	if(*iStage == 1)
	{
		const int PortCount = GetPortCount();
		for(;(*iStep)<PortCount;(*iStep)++)
		{
			int IO = GetPortIO(*iStep);
			if(Type == 1 && IO != OUTPUT && IO != INPUT && IO != INPUT_ADC && IO != OUTPUT_PWM && IO != MEMORY && IO != MEMORY_SW)
				continue;
			else if(Type == 2 && IO != OUTPUT && IO != INPUT && IO != MEMORY && IO != MEMORY_SW && IO)
				continue;
			else if(Type == 3 && IO != OUTPUT && IO != OUTPUT_PWM && IO != MEMORY && IO != MEMORY_SW)
				continue;
			else if (Type == 4 && IO != INPUT_ADC && IO != OUTPUT_PWM)
				continue;
			else if (Type == 5 && IO != OUTPUT && IO != MEMORY && IO != MEMORY_SW)
				continue;

			char* ID = GetPortID(*iStep);
			if(strprint(pBuff, nMSS, iPos, g_HtmlOpts, ID))
				return 1;
		}
		*iStep = 0;
		*iStage = 2;
	}
	if(*iStage == 2)
	{
		if(strcopy_P(pBuff, nMSS, iPos, g_HtmlOptsEnd))
			return 1;
		*iStage = 0;
	}
	return 0;
}

int PrintOptionList(char** pBuff, int* nMSS, int* iStage, int* iStep, int* iPos, char* pName, const prog_char* pOptions)
{
	if(*iStage == 0)
	{
		if(strprint(pBuff, nMSS, iPos, g_HtmlOptsBegin, pName))
			return 1;
		*iStage = 1;
	}
	if(*iStage == 1)
	{
		int i;
		for(i=0; pgm_read_byte(pOptions) != 0; i++)
		{
			if(i == *iStep)
			{
				if(strcmp_P(".",pOptions) != 0)
				{
					char Option[16];
					strcpy_P(Option,pOptions);
					if(strprint(pBuff, nMSS, iPos, g_HtmlOpts, Option))
						return 1;
				}
				(*iStep)++;
			}
			pOptions += strlen_P(pOptions)+1;
		}
		*iStep = 0;
		*iStage = 2;
	}
	if(*iStage == 2)
	{
		if(strcopy_P(pBuff, nMSS, iPos, g_HtmlOptsEnd))
			return 1;
		*iStage = 0;
	}
	return 0;
}

char* GetComboOpt(int pOpt, const prog_char* pOptions, char* pBuff)
{
	int Sel = 0;
	while(pgm_read_byte(pOptions) != 0)
	{
		if(Sel == pOpt)
		{
			strcpy_P(pBuff,pOptions);
			return pBuff;
		}
		Sel++;
		pOptions += strlen_P(pOptions)+1;
	}
	return "";
}

int GetComboVal(const char* pTest, const prog_char* pOptions)
{
	int Sel = 0;
	while(pgm_read_byte(pOptions) != 0)
	{
		if(strcmp_P(pTest,pOptions) == 0)
			return Sel;
		pOptions += strlen_P(pOptions)+1;
		Sel++;
	}
	return -1;
}

int PrintCombo(char** pBuff, int* nMSS, int* iStage, int* iPos, char* pName, char* pID, int Index, char* Sel, const char* Submit)
{
	if(Sel == NULL)
		Sel = "None";
	if(*iStage == 0)
	{
		if(strprint(pBuff, nMSS, iPos, g_HtmlCBNo, pID, Index, Sel))
			return 1;
		*iStage = 1;
	}
	if(*iStage == 1)
	{
		if(strprint(pBuff, nMSS, iPos, g_HtmlCBScript, pName, pID, Index, Sel, Submit))
			return 1;
		*iStage = 0;
	}
	return 0;
}


// HW Config
static const prog_char g_CfgHtmlForm[] =
				"<h1>Hardware Setup</h1>\r\n<form name='frmCfg'  method='POST' action='/Cfg'>\r\n";

static const prog_char g_CfgHtmlTable1[] =
				"<table border='0' cellpadding='0' cellspacing='0'>"
				"<tr><td valign='top'><table border='1px'><colgroup><col width='40px'><col width='140px'><col width='40px'></colgroup>\r\n"
				"<td>ID</td><td>Port Label</td><td>I/O</td>\r\n";

static const prog_char g_CfgHtmlTabel2[] =
				"</table></td><td >&nbsp;</td><td valign='top'><table border='1px'><colgroup><col width='40px'><col width='140px'><col width='40px'></colgroup>\r\n"
				"<td>ID</td><td>Memory Label</td><td>State</td>\r\n";

static const prog_char g_CfgHtmlLine[] =
				"<tr><td class='s%C'>%P</td><td class='s%C'><input name='L%N' value='%L' maxlength='20' size='20' type='text'></td><td class='s%C'>%X</td></tr>\r\n";
static const prog_char g_OptsIOA[] =
				"Output\0Input\0ADC\0.\0";
static const prog_char g_OptsIO[] =
				"Output\0Input\0";
static const prog_char g_OptsIOP[] =
				"Output\0Input\0.\0PWM\0";
static const prog_char g_OptsMEM[] =
				".\0.\0.\0.\0Hidden\0Visible\0";
static const prog_char g_OptsALL[] =
				"Output\0Input\0ADC\0PWM\0Hidden\0Visible\0";

static const prog_char g_HtmlEmpty[] =
				"&nbsp;";

static const prog_char g_CfgHtmlTabelEnd[] =
				"</td></tr></table></table><br><br>LCD Support:&nbsp;";

static const prog_char g_OptsLCD[] =
				"Default\0Custom\0";

static const prog_char g_CfgHtmlLines[] =
				"&nbsp;&nbsp;&nbsp;Lines: <input name='M1' value='%d' maxlength='1' size='2' type='text'>";

static const prog_char g_CfgHtmlLCD[] =
				"<br><br><table border='1px'><tr><td class='s1'>RS</td><td class='s3'>EN</td><td class='s2'>D4</td><td class='s2'>D5</td><td class='s2'>D6</td><td class='s2'>D7</td><td class='s4'>SL</td><td class='s4'>SR</td></tr>\r\n"
				"<tr><td class='s1'>%2</td><td class='s3'>%3</td><td class='s2'>%4</td><td class='s2'>%5</td><td class='s2'>%6</td><td class='s2'>%7</td><td class='s4'>%8</td><td class='s4'>%9</td></tr></table>\r\n";

static const prog_char g_CfgHtmlFormEnd[] =
				"\r\n<br><br><input name='S0' type='submit' value='Save'>\r\n"
				"</form>\r\n";

int CfgHttpData(char* pBuff, int nMSS, tTrans* Trans)
{
	char* pBuff0 = pBuff;

	if(Trans->h == 0)
	{
		if(Trans->j == 0)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderBegin))
				return pBuff-pBuff0;
			Trans->j = 1;
		}

		if(Trans->j == 1)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderStyle))
				return pBuff-pBuff0;
			Trans->j = 2;
		}

		if(Trans->j == 2)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderScript))
				return pBuff-pBuff0;
			Trans->j = 3;
		}

		if(Trans->j == 3)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "IOPorts", g_OptsIO))
				return pBuff-pBuff0;
			Trans->j = 4;
		}

		if(Trans->j == 4)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "IOPortsA", g_OptsIOA))
				return pBuff-pBuff0;
			Trans->j = 5;
		}

		if(Trans->j == 5)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "IOPortsP", g_OptsIOP))
				return pBuff-pBuff0;
			Trans->j = 6;
		}

		if(Trans->j == 6)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "IOPortsM", g_OptsMEM))
				return pBuff-pBuff0;
			Trans->j = 7;
		}

		if(Trans->j == 7)
		{
			if(GetLCDSupport() == 2)
			{
				if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "LCDPorts", 0))
					return pBuff-pBuff0;
			}
			Trans->j = 8;
		}

		if(Trans->j == 8)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderEnd))
				return pBuff-pBuff0;
			Trans->j = 0;
		}

		Trans->h = 1;
	}

	if(Trans->h == 1)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CfgHtmlForm))
			return pBuff-pBuff0;
		Trans->h = 2;
	}

	if(Trans->h == 2)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CfgHtmlTable1))
			return pBuff-pBuff0;
		Trans->h = 3;
	}

	if(Trans->h == 3)
	{
		const int CtrlHtmlLineLen = strlen_P(g_CfgHtmlLine);
		const int PortCount = GetPortCount();
		const int HWPortCount= GetHWPortCount();

		while(Trans->i<PortCount)
		{
			if(Trans->i == HWPortCount && Trans->l == 0)
			{
				if(strcopy_P(&pBuff, &nMSS, &Trans->k, g_CfgHtmlTabel2))
					return pBuff-pBuff0;
				Trans->l = 1;
			}

			while(Trans->j<CtrlHtmlLineLen)
			{
				if(strncmp_PP(g_CfgHtmlLine+Trans->j,"%N",2) == 0)
				{
					itoa(Trans->i, g_Buffer, 10);
					if(strcopy_C(&pBuff, &nMSS, &Trans->k, g_Buffer))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(strncmp_PP(g_CfgHtmlLine+Trans->j,"%P",2) == 0)
				{
					if(strcopy_C(&pBuff, &nMSS, &Trans->k, GetPortID(Trans->i)))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(strncmp_PP(g_CfgHtmlLine+Trans->j,"%L",2) == 0)
				{
					char* Label = GetPortLabel(Trans->i,0);
					if(Label)
					{
						if(strcopy_C(&pBuff, &nMSS, &Trans->k, Label))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(strncmp_PP(g_CfgHtmlLine+Trans->j,"%X",2) == 0)
				{
					int ret = 0;
					uint8_t IO = GetPortIO(Trans->i);
					if(IO >= LC_RS && IO <= LC_SR)
						ret = strcopy_C(&pBuff, &nMSS, &Trans->k, GetLCDPin(IO-LC_RS));
					else if(GetPortADC(Trans->i) != NONE)
						ret = PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->m, "IOPortsA", "F", Trans->i, GetComboOpt(IO,g_OptsIOA,Trans->Buff), "");
					else if(GetPortPWM(Trans->i) != NONE)
						ret = PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->m, "IOPortsP", "F", Trans->i, GetComboOpt(IO,g_OptsIOP,Trans->Buff), "");
					else if(GetPortMEM(Trans->i) != NONE)
						ret = PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->m, "IOPortsM", "F", Trans->i, GetComboOpt(IO,g_OptsMEM,Trans->Buff), "");
					else
						ret = PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->m, "IOPorts", "F", Trans->i, GetComboOpt(IO,g_OptsIO,Trans->Buff), "");

					if(ret)
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(strncmp_PP(g_CfgHtmlLine+Trans->j,"%C",2) == 0)
				{
					int Collor = 0;
					if(Trans->i < 8)		Collor = 1;
					else if(Trans->i < 16)	Collor = 2;
					else if(Trans->i < 24)	Collor = 3;
					else if(Trans->i < 25)	Collor = 4;
					else if(Trans->i < 33)	Collor = 5;

					itoa(Collor, g_Buffer, 10);
					if(strcopy_C(&pBuff, &nMSS, &Trans->k, g_Buffer))
						return pBuff-pBuff0;

					Trans->j+=2;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_CfgHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}
		Trans->i = 0;
		Trans->l = 0;
		Trans->h = 4;
	}

	if(Trans->h == 4)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CfgHtmlTabelEnd))
			return pBuff-pBuff0;
		Trans->h = 5;
	}

	if(Trans->h == 5)
	{
		if(Trans->j == 0)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "LCDSupport", g_OptsLCD))
				return pBuff-pBuff0;
			Trans->j = 1;
		}

		if(Trans->j == 1)
		{
			if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, "LCDSupport", "M", 0, GetLCDSupport() == 1 ? "Default" : GetLCDSupport() == 2 ? "Custom" : "None", "frmCfg"))
				return pBuff-pBuff0;
			Trans->j = 0;
		}
		Trans->h = 6;
	}

	if(Trans->h == 6)
	{
		if(GetLCDSupport())
		{
			if(strprint(&pBuff, &nMSS, &Trans->i, g_CfgHtmlLines, GetLCDLines()))
				return pBuff-pBuff0;
		}
		Trans->h = 7;
	}

	if(Trans->h == 7)
	{
		if(GetLCDSupport() == 2)
		{
			const int CfgHtmlLCDSetupLen = strlen_P(g_CfgHtmlLCD);

			while(Trans->i<CfgHtmlLCDSetupLen)
			{
				if(strncmp_PP(g_CfgHtmlLCD+Trans->i,"%",1) == 0)
				{
					strncpy_P(g_Buffer,g_CfgHtmlLCD+Trans->i+1,1);
					int Pin = atoi(g_Buffer)-2;

					if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->m, "LCDPorts", "M", Pin+2, GetLCDPinID(Pin), "frmCfg"))
						return pBuff-pBuff0;

					Trans->i+=2;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_CfgHtmlLCD+Trans->i);
					Trans->i+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->i = 0;
		}
		Trans->h = 8;
	}

	if(Trans->h == 8)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CfgHtmlFormEnd))
			return pBuff-pBuff0;
		Trans->h = 9;
	}


	if(Trans->h == 9)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlFooter))
			return pBuff-pBuff0;

		Trans->h = -1;
	}

	return pBuff-pBuff0;
}

// Option Config
static const prog_char g_MgmtHtmlForm[] =
				"<h1>Management</h1>\r\n<form name='frmMgmt'  method='POST' action='/Mgmt'>\r\n";

static const prog_char g_MgmtHtmlTable[] =
				"<table border='1'><colgroup><col width='40'><col width='40'><col width='40'><col width='40'><col width='40'><col width='110'></colgroup>\r\n"
				"<td>Input</td><td>Value</td><td>Optional</td><td>Output</td><td>Value</td><td>Position</td>\r\n";

static const prog_char g_MgmtHtmlLine[] =
				"<tr><td>%I</td><td>%J</td><td>%X</td><td>%O</td><td>%P</td><td><input value='U' type='submit' name='U%N'><input value='R' type='submit' name='R%N'><input value='D' type='submit' name='D%N'></td></tr>\r\n";

static const prog_char g_OptsBitV[] =
				"Lo\0Hi\0Lo.S\0Hi.S\0";
static const prog_char g_OptsBinV[] =
				"Lo\0Hi\0AND\0NOT.AND\0AND.NOT\0NOR\0OR\0NOT.OR\0OR.NOT\0NAND\0XOR\0XNOR\0";
static const prog_char g_OptsAdcV[] =
				"EQ.v\0GT.v\0LT.v\0NE.v\0EQ.c\0GT.c\0LT.c\0NE.c\0";

static const prog_char g_MgmtHtmlEB[] =
				"<input name='%s%d' onblur = 'frmMgmt.submit();' value='%d' maxlength='10' size='5' type='text'>";

static const prog_char g_MgmtHtmlFormEnd[] =
				"<tr><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td><td><input value='A' type='submit' name='A00'></td><tr></table>\r\n"
				"<br><input name='S1' type='submit' value='Update'>&nbsp;<input name='S1' type='submit' value='Save'></form>\r\n";

int MgmtHttpData(char* pBuff, int nMSS, tTrans* Trans)
{
	char* pBuff0 = pBuff;

	if(Trans->h == 0)
	{
		if(Trans->j == 0)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderBegin))
				return pBuff-pBuff0;
			Trans->j = 1;
		}

		if(Trans->j == 1)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderScript))
				return pBuff-pBuff0;
			Trans->j = 2;
		}

		if(Trans->j == 2)
		{
			if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "IOPorts", 1))
				return pBuff-pBuff0;
			Trans->j = 3;
		}

		if(Trans->j == 3)
		{
			if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "IOPorts2", 2))
				return pBuff-pBuff0;
			Trans->j = 4;
		}

		if(Trans->j == 4)
		{
			if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "OPorts", 3))
				return pBuff-pBuff0;
			Trans->j = 5;
		}

		if(Trans->j == 5)
		{
			if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "APorts", 4))
				return pBuff-pBuff0;
			Trans->j = 6;
		}

		if(Trans->j == 6)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "BitV", g_OptsBitV))
				return pBuff-pBuff0;
			Trans->j = 7;
		}

		if(Trans->j == 7)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "BinV", g_OptsBinV))
				return pBuff-pBuff0;
			Trans->j = 8;
		}

		if(Trans->j == 8)
		{
			if(PrintOptionList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "AdcV", g_OptsAdcV))
				return pBuff-pBuff0;
			Trans->j = 9;
		}

		if(Trans->j == 9)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderEnd))
				return pBuff-pBuff0;
			Trans->j = 0;
		}
		Trans->h = 1;
	}

	if(Trans->h == 1)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_MgmtHtmlForm))
			return pBuff-pBuff0;
		Trans->h = 2;
	}

	if(Trans->h == 2)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_MgmtHtmlTable))
			return pBuff-pBuff0;
		Trans->h = 3;
	}

	if(Trans->h == 3)
	{
		const int MgmtHtmlLineLen = strlen_P(g_MgmtHtmlLine);
		const int OptCount = GetOptCount();

		while(Trans->i<OptCount)
		{
			while(Trans->j<MgmtHtmlLineLen)
			{
				if(strncmp_PP(g_MgmtHtmlLine+Trans->j,"%N",2) == 0)
				{
					itoa(Trans->i, g_Buffer, 10);
					if(strcopy_C(&pBuff, &nMSS, &Trans->k, g_Buffer))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(strncmp_PP(g_MgmtHtmlLine+Trans->j,"%I",2) == 0)
				{
					if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, "IOPorts", "I", Trans->i, GetPortID(GetOptIn(Trans->i)), "frmMgmt"))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(strncmp_PP(g_MgmtHtmlLine+Trans->j,"%J",2) == 0)
				{
					switch(GetPortIO(GetOptIn(Trans->i)))
					{
					case OUTPUT:
					case INPUT:
					case MEMORY:
					case MEMORY_SW:
						if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, "BinV", "J", Trans->i, GetComboOpt(GetOptInVal(Trans->i),g_OptsBinV,Trans->Buff), "frmMgmt"))
							return pBuff-pBuff0;
						break;
					case INPUT_ADC:
					case OUTPUT_PWM:
						if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, "AdcV", "J", Trans->i, GetComboOpt(GetOptInVal(Trans->i),g_OptsAdcV,Trans->Buff), "frmMgmt"))
							return pBuff-pBuff0;
						break;
					default:
						if(strcopy_P(&pBuff, &nMSS, &Trans->k, g_HtmlEmpty))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(strncmp_PP(g_MgmtHtmlLine+Trans->j,"%X",2) == 0)
				{
					switch(GetPortIO(GetOptIn(Trans->i)))
					{
					case OUTPUT:
					case INPUT:
					case MEMORY:
					case MEMORY_SW:
						if(GetOptInVal(Trans->i) < AND)
						{
							int Aux = GetOptAux(Trans->i);
							if(Aux == 0xFFFF)
								Aux = 0;
							if(strprint(&pBuff, &nMSS, &Trans->k, g_MgmtHtmlEB,"E",Trans->i,Aux))
								return pBuff-pBuff0;
						}
						else
						{
							if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, "IOPorts2", "E", Trans->i, GetPortID(GetOptAux(Trans->i)), "frmMgmt"))
								return pBuff-pBuff0;
						}
						break;
					case INPUT_ADC:
					case OUTPUT_PWM:
						if(GetOptInVal(Trans->i) < EQ_CHN)
						{
							int Aux = GetOptAux(Trans->i);
							if(Aux == 0xFFFF)
								Aux = 0;
							if(strprint(&pBuff, &nMSS, &Trans->k, g_MgmtHtmlEB,"E",Trans->i,Aux))
								return pBuff-pBuff0;
						}
						else
						{
							if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, "APorts", "E", Trans->i, GetPortID(GetOptAux(Trans->i)), "frmMgmt"))
								return pBuff-pBuff0;
						}
						break;
					default:
						if(strcopy_P(&pBuff, &nMSS, &Trans->k, g_HtmlEmpty))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(strncmp_PP(g_MgmtHtmlLine+Trans->j,"%O",2) == 0)
				{
					if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, "OPorts", "O", Trans->i, GetPortID(GetOptOut(Trans->i)), "frmMgmt"))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(strncmp_PP(g_MgmtHtmlLine+Trans->j,"%P",2) == 0)
				{
					switch(GetPortIO(GetOptOut(Trans->i)))
					{
					case OUTPUT:
					case MEMORY:
					case MEMORY_SW:
						if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, "BitV", "P", Trans->i, GetComboOpt(GetOptOutVal(Trans->i),g_OptsBitV,Trans->Buff), "frmMgmt"))
							return pBuff-pBuff0;
						break;
					case OUTPUT_PWM:
						if(strprint(&pBuff, &nMSS, &Trans->k, g_MgmtHtmlEB,"P",Trans->i,GetOptOutVal(Trans->i)))
							return pBuff-pBuff0;
						break;
					default:
						if(strcopy_P(&pBuff, &nMSS, &Trans->k, g_HtmlEmpty))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_MgmtHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}
		Trans->i = 0;
		Trans->h = 4;
	}

	if(Trans->h == 4)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_MgmtHtmlFormEnd))
			return pBuff-pBuff0;
		Trans->h = 5;
	}

	if(Trans->h == 5)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlFooter))
			return pBuff-pBuff0;
		Trans->h = -1;
	}

	return pBuff-pBuff0;
}

// Gui
static const prog_char g_CtrlHtmlTable[] =
				"<h1>Controll Pannel</h1>\r\n"
				"<form form name='frmCtrl' method='POST' action='/'>\r\n"
				"<table border='0' cellpadding='0' cellspacing='0'><tr>\r\n"
				"<td valign='top'>\r\n"
				"<table border='1'><colgroup><col width='140'><col width='40'></colgroup>\r\n";

static const prog_char g_CtrlHtmlLineTop[] =
				"<td>%s</td><td>%s</td>\r\n";

static const prog_char g_CtrlHtmlLine[] =
				"<tr><td>%L</td><td height='29'>%X</td></tr>\r\n";

static const prog_char g_CtrlHtmlClick[] =
				"<input value='%s' type='submit' name='%s%d'>";

static const prog_char g_CtrlHtmlEB[] =
				"<input name='%s%d' onblur = 'frmCtrl.submit();' value='%d' maxlength='10' size='5' type='text'>\r\n";

static const prog_char g_CtrlHtmlSpacer[] =
				"</table></td><td valign='top'>&nbsp;</td><td valign='top'>\r\n"
				"<table border='1'><colgroup><col width='140'><col width='40'></colgroup>\r\n";


static const prog_char g_CtrlHtmlTableEnd[] =
				"</table></form></td>\r\n"
				"</tr></table>\r\n"
				"<p><a href='/Mgmt'>Management</a>&nbsp;<a href='/Cfg'>Configuration</a>&nbsp;<a href='/Lcd'>Lcd</a></p>";

int CtrlHttpData(char* pBuff, int nMSS, tTrans* Trans)
{
	char* pBuff0 = pBuff;

	if(Trans->h == 0)
	{
		if(Trans->j == 0)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderBegin))
				return pBuff-pBuff0;
			Trans->j = 1;
		}

		if(Trans->j == 1)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderEnd))
				return pBuff-pBuff0;
			Trans->j = 0;
		}
		Trans->h = 1;
	}

	if(Trans->h == 1)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlTable))
			return pBuff-pBuff0;
		Trans->h = 2;
	}

	const int CtrlHtmlLineLen = strlen_P(g_CtrlHtmlLine);

	if(Trans->h == 2)
	{
		if(strprint(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlLineTop,"Input","Value"))
			return pBuff-pBuff0;
		Trans->h = 3;
	}

	const int HWPortCount= GetHWPortCount();

	if(Trans->h == 3)
	{
		while(Trans->i<HWPortCount)
		{
			uint8_t IO = GetPortIO(Trans->i);
			if((IO != INPUT && IO != INPUT_ADC) || (IO >= LC_RS && IO <= LC_SR))
			{
				Trans->i++;
				continue;
			}

			while(Trans->j<CtrlHtmlLineLen)
			{
				if(strncmp_PP(g_CtrlHtmlLine+Trans->j,"%L",2) == 0)
				{
					char* Label = GetPortLabel(Trans->i,1);
					if(Label)
					{
						if(strcopy_C(&pBuff, &nMSS, &Trans->k, Label))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(strncmp_PP(g_CtrlHtmlLine+Trans->j,"%X",2) == 0)
				{
					switch(GetPortIO(Trans->i))
					{
					case INPUT:
						if(strcopy_C(&pBuff, &nMSS, &Trans->k, GetBit(Trans->i) ? "Hi" : "Lo"))
						//if(strprint(&pBuff, &nMSS, &Trans->k, g_CtrlHtmlClick,GetBit(Trans->i) ? "Hi" : "Lo", "V",Trans->i))
							return pBuff-pBuff0;
						break;
					case INPUT_ADC:
						itoa(GetVal(Trans->i), g_Buffer, 10);
						if(strcopy_C(&pBuff, &nMSS, &Trans->k, g_Buffer))
						//if(strprint(&pBuff, &nMSS, &Trans->k, g_CtrlHtmlEB,"V",Trans->i,GetVal(Trans->i)))
							return pBuff-pBuff0;
						break;
					}
					Trans->j+=2;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_CtrlHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}

		Trans->i = 0;
		Trans->h = 4;
	}

	if(Trans->h == 4)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlSpacer))
			return pBuff-pBuff0;
		Trans->h = 5;
	}

	if(Trans->h == 5)
	{
		if(strprint(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlLineTop,"Output","Value"))
			return pBuff-pBuff0;
		Trans->h = 6;
	}

	if(Trans->h == 6)
	{
		while(Trans->i<HWPortCount)
		{
			uint8_t IO = GetPortIO(Trans->i);
			if((IO != OUTPUT && IO != OUTPUT_PWM) || (IO >= LC_RS && IO <= LC_SR))
			{
				Trans->i++;
				continue;
			}

			while(Trans->j<CtrlHtmlLineLen)
			{
				if(strncmp_PP(g_CtrlHtmlLine+Trans->j,"%L",2) == 0)
				{
					char* Label = GetPortLabel(Trans->i,1);
					if(Label)
					{
						if(strcopy_C(&pBuff, &nMSS, &Trans->k, Label))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(strncmp_PP(g_CtrlHtmlLine+Trans->j,"%X",2) == 0)
				{
					switch(GetPortIO(Trans->i))
					{
					case OUTPUT:
						if(strprint(&pBuff, &nMSS, &Trans->k, g_CtrlHtmlClick,GetBit(Trans->i) ? "Hi" : "Lo", "V",Trans->i))
							return pBuff-pBuff0;
						break;
					case OUTPUT_PWM:
						if(strprint(&pBuff, &nMSS, &Trans->k, g_CtrlHtmlEB,"V",Trans->i,GetVal(Trans->i)))
							return pBuff-pBuff0;
						break;
					}
					Trans->j+=2;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_CtrlHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}
		Trans->i = 0;
		Trans->h = 7;
	}

	const int MEMPortCount = GetPortCount()-HWPortCount;
	if(Trans->h == 7)
	{
		uint8_t Mem = 0;
		int i;
		for(i=0; i<MEMPortCount;i++)
		{
			if(GetPortIO(HWPortCount+i) == MEMORY_SW)
			{
				Mem = 1;
				break;
			}
		}

		if(Mem)
			Trans->h = 8;
		else
			Trans->h = 11;
	}

	if(Trans->h == 8)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlSpacer))
			return pBuff-pBuff0;
		Trans->h = 9;
	}

	if(Trans->h == 9)
	{
		if(strprint(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlLineTop,"Memory","Value"))
			return pBuff-pBuff0;
		Trans->h = 10;
	}

	if(Trans->h == 10)
	{
		while(Trans->i<MEMPortCount)
		{
			if(GetPortIO(Trans->i+HWPortCount) != MEMORY_SW)
			{
				Trans->i++;
				continue;
			}

			while(Trans->j<CtrlHtmlLineLen)
			{
				if(strncmp_PP(g_CtrlHtmlLine+Trans->j,"%L",2) == 0)
				{
					char* Label = GetPortLabel(Trans->i+HWPortCount,1);
					if(Label)
					{
						if(strcopy_C(&pBuff, &nMSS, &Trans->k, Label))
							return pBuff-pBuff0;
					}
					Trans->j+=2;
				}
				else if(strncmp_PP(g_CtrlHtmlLine+Trans->j,"%X",2) == 0)
				{
					if(strprint(&pBuff, &nMSS, &Trans->k, g_CtrlHtmlClick,GetBit(Trans->i+HWPortCount) ? "Hi" : "Lo","V",Trans->i+HWPortCount))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_CtrlHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}
		Trans->i = 0;
		Trans->h = 11;
	}


	if(Trans->h == 11)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_CtrlHtmlTableEnd))
			return pBuff-pBuff0;
		Trans->h = 12;
	}

	if(Trans->h == 12)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlFooter))
			return pBuff-pBuff0;
		Trans->h = -1;
	}

	return pBuff-pBuff0;
}

// LCD Options
static const prog_char g_LcdHtmlTable[] =
				"<h1>LCD</h1>\r\n"
				"<form form name='frmLcd' method='POST' action='/Lcd'>\r\n"
				"<table border='1'>\r\n";

static const prog_char g_LcdHtmlLine[] =
				"<tr>%S<td>%X</td><td><input value='T' type='submit' name='T%N'><input value='K' type='submit' name='K%N'><input value='B' type='submit' name='B%N'></td></tr>\r\n";

static const prog_char g_LcdHtmlScr[] =
				"<td><input name='%s%d' value='%s' maxlength='40' size='20' type='text'></td>";

static const prog_char g_LcdHtmlAdd[] =
				"<tr><td colspan='%d'>Auto Switch Screans Delay: &nbsp;<input name='Q0' onblur = 'frmLcd.submit();' value='%d' maxlength='10' size='5' type='text'></td>";

static const prog_char g_LcdHtmlTableEnd[] =
				"<td><input value='N' type='submit' name='N00'></td><tr>\r\n</table>\r\n"
				"<br><input name='S2' type='submit' value='Save'></form>\r\n";

int LcdHttpData(char* pBuff, int nMSS, tTrans* Trans)
{
	char* pBuff0 = pBuff;

	if(Trans->h == 0)
	{
		if(Trans->j == 0)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderBegin))
				return pBuff-pBuff0;
			Trans->j = 1;
		}

		if(Trans->j ==1)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderScript))
				return pBuff-pBuff0;
			Trans->j = 2;
		}

		if(Trans->j == 2)
		{
			if(PrintPortsList(&pBuff, &nMSS, &Trans->k, &Trans->l, &Trans->m, "OPorts", 5))
				return pBuff-pBuff0;
			Trans->j = 3;
		}

		if(Trans->j == 3)
		{
			if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlHeaderEnd))
				return pBuff-pBuff0;
			Trans->j = 0;
		}
		Trans->h = 1;
	}

	if(Trans->h == 1)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_LcdHtmlTable))
			return pBuff-pBuff0;
		Trans->h = 2;
	}

	const int LineCount = GetLCDLines();
	const int ScrCount = GetScrCount();
	if(Trans->h == 2)
	{
		const int LcdHtmlLineLen = strlen_P(g_LcdHtmlLine);

		while(Trans->i<ScrCount)
		{
			while(Trans->j<LcdHtmlLineLen)
			{
				if(strncmp_PP(g_LcdHtmlLine+Trans->j,"%N",2) == 0)
				{
					itoa(Trans->i, g_Buffer, 10);
					if(strcopy_C(&pBuff, &nMSS, &Trans->k, g_Buffer))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else if(strncmp_PP(g_LcdHtmlLine+Trans->j,"%S",2) == 0)
				{
					while(Trans->l<LineCount)
					{
						char SCR[2] = " ";
						SCR[0] = 'W'+Trans->l;
						if(strprint(&pBuff, &nMSS, &Trans->k, g_LcdHtmlScr,SCR,Trans->i,GetScr(Trans->i,Trans->l)))
							return pBuff-pBuff0;
						Trans->l++;
					}
					Trans->l = 0;
					Trans->j+=2;
				}
				else if(strncmp_PP(g_LcdHtmlLine+Trans->j,"%X",2) == 0)
				{
					if(PrintCombo(&pBuff, &nMSS, &Trans->k, &Trans->l, "OPorts", "H", Trans->i, GetPortID(GetScrAction(Trans->i)), ""))
						return pBuff-pBuff0;
					Trans->j+=2;
				}
				else
				{
					*pBuff++ = pgm_read_byte(g_LcdHtmlLine+Trans->j);
					Trans->j+=1;
					if(!--nMSS)
						return pBuff-pBuff0;
				}
			}
			Trans->j = 0;
			Trans->i++;
		}
		Trans->i = 0;
		Trans->h = 3;
	}

	if(Trans->h == 3)
	{
		if(strprint(&pBuff, &nMSS, &Trans->i, g_LcdHtmlAdd, LineCount+1, GetLCDAuto()))
			return pBuff-pBuff0;
		Trans->h = 4;
	}

	if(Trans->h == 4)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_LcdHtmlTableEnd))
			return pBuff-pBuff0;
		Trans->h = 5;
	}

	if(Trans->h == 5)
	{
		if(strcopy_P(&pBuff, &nMSS, &Trans->i, g_HtmlFooter))
			return pBuff-pBuff0;
		Trans->h = -1;
	}

	return pBuff-pBuff0;
}

int MakeHttpData(char* pBuff, int nMSS, tTrans* Trans)
{
	if(Trans->Index == 0)
		return CtrlHttpData(pBuff, nMSS, Trans);
	else if(Trans->Index == 1)
		return CfgHttpData(pBuff, nMSS, Trans);
	else if(Trans->Index == 2)
		return MgmtHttpData(pBuff, nMSS, Trans);
	else if(Trans->Index == 3)
		return LcdHttpData(pBuff, nMSS, Trans);
	return 0;
}

// post Handling
void PostValue(char* pBuff, int* i, int* n)
{
	char OP = *pBuff++;
	int ID = atoi(pBuff);
	while(*pBuff++)
	{
		if(*pBuff == '=')
		{
			pBuff++;
			break;
		}
	}

	int j,k,l;
	for(j=k=0;pBuff[j] != 0;j++)
	{
		if(pBuff[j] == '+')
			pBuff[k] = ' ';
		else if(pBuff[j] == '%')
		{
			sscanf(pBuff+j+1,"%2x",&l);
			pBuff[k] = (char)l;
			j+=2;
		}
		else if(k!=j)
			pBuff[k] = pBuff[j];
		k++;
	}
	pBuff[k] = 0;

	switch(OP)
	{
	case 'L':
		SetPortLabel(ID, pBuff);
		break;
	case 'F':
		SetPortIO(ID, GetComboVal(pBuff, g_OptsALL));
		break;

	case 'M':
		switch(ID)
		{
			case 0: // Support
				SetLCDSupport(GetComboVal(pBuff, g_OptsLCD)+1);
			break;
			case 1: // Lines
				SetLCDLines(atoi(pBuff));
			break;
			default:
				SetLCDPin(ID-2,GetPort(pBuff));
		}

	case 'A':
		AddOption(0);
		break;
	case 'U':
		*i=1;
		*n = ID;
		break;
	case 'R':
		*i=2;
		*n = ID;
		break;
	case 'D':
		*i=3;
		*n = ID;
		break;

	case 'I':
		SetOptIn(ID, GetPort(pBuff));
		break;
	case 'J':
		switch(GetPortIO(GetOptIn(ID)))
		{
		case OUTPUT:
		case INPUT:
		case MEMORY:
		case MEMORY_SW:
			SetOptInVal(ID, GetComboVal(pBuff, g_OptsBinV));
			break;
		case INPUT_ADC:
		case OUTPUT_PWM:
			SetOptInVal(ID, GetComboVal(pBuff, g_OptsAdcV));
			break;
		}
		break;
	case 'E':
		switch(GetPortIO(GetOptIn(ID)))
		{
		case OUTPUT:
		case INPUT:
		case MEMORY:
		case MEMORY_SW:
			if(GetOptInVal(ID) < AND)
				SetOptAux(ID,atoi(pBuff));
			else
				SetOptAux(ID,GetPort(pBuff));
			break;
		case INPUT_ADC:
		case OUTPUT_PWM:
			if(GetOptInVal(ID) < EQ_CHN)
				SetOptAux(ID,atoi(pBuff));
			else
				SetOptAux(ID,GetPort(pBuff));
			break;
		}
		break;
	case 'O':
		SetOptOut(ID, GetPort(pBuff));
		break;
	case 'P':
		switch(GetPortIO(GetOptOut(ID)))
		{
		case MEMORY:
		case MEMORY_SW:
		case OUTPUT:
			SetOptOutVal(ID,GetComboVal(pBuff, g_OptsBitV));
			break;
		case OUTPUT_PWM:
			SetOptOutVal(ID,atoi(pBuff));
			break;
		}
		break;

	case 'V':
		switch(GetPortIO(ID))
		{
		case INPUT:
		case OUTPUT:
		case MEMORY:
		case MEMORY_SW:
			if(strcmp(pBuff, "Hi") == 0)
				SetBit(ID,0);
			if(strcmp(pBuff, "Lo") == 0)
				SetBit(ID,1);
			break;
		case INPUT_ADC:
		case OUTPUT_PWM:
			SetVal(ID,atoi(pBuff));
			break;
		}
		break;

	case 'N':
		AddScreen();
		break;
	case 'T':
		*i=4;
		*n = ID;
		break;
	case 'K':
		*i=5;
		*n = ID;
		break;
	case 'B':
		*i=6;
		*n = ID;
		break;

	case 'W':
		SetScr(ID,0,pBuff);
		break;
	case 'X':
		SetScr(ID,1,pBuff);
		break;
	case 'Y':
		SetScr(ID,2,pBuff);
		break;
	case 'Z':
		SetScr(ID,3,pBuff);
		break;

	case 'H':
		SetScrAction(ID,GetPort(pBuff));
		break;

	case 'Q':
		SetLCDAuto(atoi(pBuff));
		break;

	case 'S':
		if(strcmp_P(pBuff,PSTR("Save")) == 0)
		{
			*i=7;
			*n = ID;
		}
		break;
	}
}

// Request Processing

uint8_t strfind(char** pBuff, int *iLen, int* iPos, const char* pStr)
{
	while(*iLen)
	{
		(*iLen)--;
		if(**pBuff == pStr[*iPos])
		{
			(*iPos)++;
			if(pStr[*iPos] == '\0')
			{
				(*pBuff)++;
				*iPos = 0;
				return 0;
			}
		}
		else
			*iPos = 0;
		(*pBuff)++;
	}
	return 1;
}

uint8_t strcopyto(char* pStr, int* iPos, char** pBuff, int *iLen, char* pEnd, int *iEnd, int *iLen2)
{
	while(*iLen)
	{
		(*iLen)--;
		if(iLen2) (*iLen2)--;
		pStr[*iPos] = **pBuff;
		if(**pBuff == pEnd[*iEnd])
		{
			(*iEnd)++;
			if(pEnd[*iEnd] == '\0')
			{
				*iEnd = 0;
				(*pBuff)++;
				return 0;
			}
		}
		else
			*iEnd = 0;
		(*pBuff)++;
		(*iPos)++;
	}
	if(iLen2 && *iLen2 == 0)
		return 0;
	return 1;
}

int ParseHeader(char* pBuff, int iLen, tTrans* Trans)
{
	if(Trans->h == 0)
	{
		if(Trans->m < 4)
		{
			int togo = iLen <= (4-Trans->m) ? iLen : (4-Trans->m);
			strncpy(Trans->Buff+Trans->m,pBuff,togo);
			Trans->m += togo;
			pBuff += togo;
			iLen -= togo;
			if(Trans->m < 4)
				return 1;
		}

		if(strncmp(Trans->Buff,"GET",3) == 0)
		{
			Trans->m = 0;
			Trans->h = 2;
		}
		else if(strncmp(Trans->Buff,"POST",4) == 0)
		{
			if(iLen == 0)
				return 1;
			pBuff += 1;
			iLen -= 1; // skip the space
			Trans->m = 0;
			Trans->h = 1;
		}
		else
		{
			Trans->m = 0;
			Trans->h = 10;
			return 1;
		}
	}

	if(Trans->h == 1 || Trans->h == 2)
	{
		if(strcopyto(Trans->Buff, &Trans->m, &pBuff, &iLen, "HTTP", &Trans->l, NULL))
			return 1;
		Trans->m = 0;

		if(strncmp(Trans->Buff,"/ ",2) == 0)
		{
			Trans->Index = 0;
			Trans->h += 2;
		}
		else if(strncmp(Trans->Buff,"/Cfg ",5) == 0)
		{
			Trans->Index = 1;
			Trans->h += 2;
		}
		else if(strncmp(Trans->Buff,"/Mgmt ",6) == 0)
		{
			Trans->Index = 2;
			Trans->h += 2;
		}
		else if(strncmp(Trans->Buff,"/Lcd ",5) == 0)
		{
			Trans->Index = 3;
			Trans->h += 2;
		}
		else if(strncmp(Trans->Buff,"/Reboot ",8) == 0)
		{
			Reboot();
		}
		else
		{
			Trans->h = 10;
			return 1;
		}
	}

	if(Trans->h == 3)
	{

		if(Trans->k == 0)
		{
			if(strfind(&pBuff, &iLen, &Trans->m,"Content-Length:"))
				return 1;
			Trans->k = 1;
		}

		if(Trans->k == 1)
		{
			if(strcopyto(Trans->Buff, &Trans->m, &pBuff, &iLen, "\r\n", &Trans->l, NULL))
				return 1;

			Trans->Buff[Trans->m] = '\0';
			Trans->m = 2; // mark that we already have a \r\n pair!
			Trans->j = atoi(Trans->Buff);

			Trans->k = 2;
		}

		if(Trans->k == 2)
		{
			if(strfind(&pBuff, &iLen, &Trans->m,"\r\n\r\n"))
				return 1;
			Trans->k = 3;
		}

		if(Trans->k == 3)
		{
			while(Trans->j)
			{
				if(strcopyto(Trans->Buff, &Trans->m, &pBuff, &iLen, "&", &Trans->l, &Trans->j))
					return 1;
				Trans->Buff[Trans->m] = '\0';
				Trans->m = 0;
				PostValue(Trans->Buff,&Trans->i,&Trans->n);
			}
			Trans->k = 0;
		}

		if(Trans->i)
		{
			switch(Trans->i)
			{
			case 1:
				UpOption(Trans->n);
				break;
			case 2:
				RemoveOption(Trans->n);
				break;
			case 3:
				DownOption(Trans->n);
				break;
			case 4:
				UpScreen(Trans->n);
				break;
			case 5:
				RemoveScreen(Trans->n);
				break;
			case 6:
				DownScreen(Trans->n);
				break;
			case 7: // Save
				Save(Trans->n);
			}

			Trans->i = 0;
			Trans->n = 0;
		}

		Trans->h = 0;
	}
	else if(Trans->h == 4)
		Trans->h = 0;

	return 0;
}

