/*
 * Options.c
 *
 *      Author: Trax
 */

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Options.h"
#include "Ports.h"

tOption* g_Option = NULL;
int	g_OptionCnt = 0;

int GetOptCount(){
	return g_OptionCnt;
}

int GetOptIn(int i){
	if(i >= GetOptCount())
		return 0;
	return g_Option[i].Input;
}
void SetOptIn(int i, int Val){
	if(i >= GetOptCount())
		return;
	g_Option[i].Input = Val;
}
int GetOptInVal(int i){
	if(i >= GetOptCount())
		return 0;
	return g_Option[i].InValue;
}
void SetOptInVal(int i, int Val){
	if(i >= GetOptCount())
		return;
	g_Option[i].InValue = Val;
}
int GetOptAux(int i){
	if(i >= GetOptCount())
		return 0;
	return g_Option[i].Aux;
}
void SetOptAux(int i, int Val){
	if(i >= GetOptCount())
		return;
	g_Option[i].Aux = Val;
}
int GetOptOut(int i){
	if(i >= GetOptCount())
		return 0;
	return g_Option[i].Output;
}
void SetOptOut(int i, int Val){
	if(i >= GetOptCount())
		return;
	g_Option[i].Output = Val;
}
int GetOptOutVal(int i){
	if(i >= GetOptCount())
		return 0;
	return g_Option[i].OutValue;
}
void SetOptOutVal(int i, int Val){
	if(i >= GetOptCount())
		return;
	g_Option[i].OutValue = Val;
}

void AddOption(uint8_t* Opt)
{
	if(g_Option)
		g_Option = (tOption*) realloc(g_Option, sizeof(tOption)*(g_OptionCnt+1));
	else
		g_Option = (tOption*) malloc(sizeof(tOption));


	if(Opt)
		memcpy(&g_Option[g_OptionCnt],Opt,sizeof(tOption));
	else
		memset(&g_Option[g_OptionCnt],0xff,sizeof(tOption));

	g_OptionCnt++;
}

void UpOption(int i)
{
	if(i >= g_OptionCnt || i == 0)
		return;

	tOption temp;
	memcpy(&temp,&g_Option[i-1],sizeof(tOption));
	memcpy(&g_Option[i-1],&g_Option[i],sizeof(tOption));
	memcpy(&g_Option[i],&temp,sizeof(tOption));
}

void RemoveOption(int i)
{
	if(i >= g_OptionCnt)
		return;
	if(g_OptionCnt == 1)
	{
		free(g_Option);
		g_Option = NULL;
	}
	else
	{
		tOption* l_Option = (tOption*)malloc(sizeof(tOption)*(g_OptionCnt-1));

		if(i > 0)
			memcpy(l_Option,g_Option,i*sizeof(tOption));
		if(i+1 < g_OptionCnt)
			memcpy(&l_Option[i],&g_Option[i+1],(g_OptionCnt-(i+1))*sizeof(tOption));

		free(g_Option);
		g_Option = l_Option;
	}
	g_OptionCnt--;
}

void DownOption(int i)
{
	if(i >= g_OptionCnt || i == g_OptionCnt -1)
		return;

	tOption temp;
	memcpy(&temp,&g_Option[i+1],sizeof(tOption));
	memcpy(&g_Option[i+1],&g_Option[i],sizeof(tOption));
	memcpy(&g_Option[i],&temp,sizeof(tOption));
}


////////////////////////////////////////////////////////////

typedef struct
{
	uint16_t Option;
	uint16_t Delay;

	void* Next;
} tDelay;

tDelay* g_Delay = NULL;

uint8_t Delay(uint16_t Option, uint16_t Value)
{
	tDelay* cDelay = g_Delay;
	tDelay* pDelay = NULL;
	while(cDelay)
	{
		if(cDelay->Option == Option)
		{
			if(Value)
			{
				if(cDelay->Delay)
				{
					cDelay->Delay--;
					return 0;
				}
				return 1;
			}
			else
			{
				if(pDelay)
					pDelay->Next = cDelay->Next;
				else
					g_Delay = (tDelay*)cDelay->Next;
				free (cDelay);
				return 1;
			}
		}
		pDelay = cDelay;
		cDelay = (tDelay*)cDelay->Next;
	}

	if(Value)
	{
		cDelay = (tDelay*)malloc(sizeof(tDelay));
		cDelay->Delay = Value;
		cDelay->Option = Option;
		cDelay->Next = NULL;

		if(pDelay)
			pDelay->Next = cDelay;
		else
			g_Delay = cDelay;
		return 0;
	}
	else
	{
		return 1;
	}
}

//Process Options
void MgmtDo()
{
	int i;
	for(i=0; i<g_OptionCnt;i++)
	{
		uint8_t Result = 0;
		uint16_t Wait = 0;
		switch(GetPortIO(g_Option[i].Input))
		{
		case OUTPUT:
		case INPUT:
		case MEMORY:
		case MEMORY_SW:
			{
				uint8_t Input = GetBit(g_Option[i].Input);
				uint16_t AuxInput = 0;
				if(g_Option[i].InValue >= AND)
					AuxInput = GetBit(g_Option[i].Aux);
				switch(g_Option[i].InValue)
				{
				case LO:
				case HI:
					Wait = g_Option[i].Aux;
					Result = Input == g_Option[i].InValue;
					break;
				case AND:
					Result = Input && AuxInput;
					break;
				case NOT_AND:
					Result = !Input && AuxInput;
					break;
				case AND_NOT:
					Result = Input && !AuxInput;
					break;
				case NOR:
					Result = !Input && !AuxInput;
					break;
				case OR:
					Result = Input || AuxInput;
					break;
				case NOT_OR:
					Result = !Input || AuxInput;
					break;
				case OR_NOT:
					Result = Input || !AuxInput;
					break;
				case NAND:
					Result = !Input || !AuxInput;
					break;
				case XOR:
					Result = Input ^ AuxInput;
					break;
				case XNOR:
					Result = Input ^ !AuxInput;
					break;
				}
				break;
			}
		case INPUT_ADC:
		case OUTPUT_PWM:
			{
				uint16_t Value = GetVal(g_Option[i].Input);
				uint16_t AuxValue;
				if(g_Option[i].InValue < EQ_CHN)
					AuxValue = g_Option[i].Aux;
				else
					AuxValue = GetVal(g_Option[i].Input);
				switch(g_Option[i].InValue)
				{
				case EQ_VAL:
				case EQ_CHN:
					Result = Value == AuxValue;
					break;
				case GT_VAL:
				case GT_CHN:
					Result = Value > AuxValue;
					break;
				case LT_VAL:
				case LT_CHN:
					Result = Value < AuxValue;
					break;
				case NE_VAL:
				case NE_CHN:
					Result = Value != AuxValue;
					break;
				}
				break;
			}
		default:
			continue;
		}

		if(Wait > 0 && !Delay(i,Result ? Wait : 0))
			continue;


		switch(GetPortIO(g_Option[i].Output))
		{
		case OUTPUT:
		case MEMORY:
		case MEMORY_SW:
			switch(g_Option[i].OutValue)
			{
			case LO_S:
				if(Result)
					SetBit(g_Option[i].Output, LO);
				break;
			case HI_S:
				if(Result)
					SetBit(g_Option[i].Output, HI);
				break;
			case LO:
				if(Result)
					SetBit(g_Option[i].Output, LO);
				else
					SetBit(g_Option[i].Output, HI);
				break;
			case HI:
				if(Result)
					SetBit(g_Option[i].Output, HI);
				else
					SetBit(g_Option[i].Output, LO);
				break;
			}

			break;
		case OUTPUT_PWM:
			if(Result)
				SetVal(g_Option[i].Output, g_Option[i].OutValue);
			break;
		default:
			continue;
		}
	}
}
